{
MIDI files decoding and playing code for Win32
----------------------------------------------

Based on MSVC++ source code of TMIDI by Tom Grandgent

Rewrited on Object Pascal for Delphi 7 by Sergey Bulba

(c)2006 S.V.Bulba
http://bulba.at.kz/
vorobey@mail.khstu.ru

This is first edition of rewriting and removing extra code
from original TMIDI. Only playing, seeking and visualisation
code is used. Must play all MIDI files of types from 0 to 2
(.MID and .RMI). Uses the lowest level of Win32 Multimedia
System
}

unit Midi;

interface

uses Windows,MMSystem,Sysutils,StdCtrls;

const
 MThd = $6468544D;
 MTrk = $6B72544D;
 RIFF = $46464952;
 MIDI_STANDARD_NONE=1;
 MIDI_STANDARD_GM=2;
 MIDI_STANDARD_GS=4;
 MIDI_STANDARD_XG=8;
 MIDI_STANDARD_MT32=16;
 MAX_MIDI_WAIT=1000.0;	// Max # of milliseconds to sleep in main playback loop
 MAX_MIDI_TRACKS=512;		// Max # of tracks DMIDI will read from a MIDI file
 MAX_PITCH_BEND=$2000;	// Max value for pitch bends (this is a MIDI constant, do not change!)

type
 PByteArray = ^TByteArray;
 TByteArray = packed array[0..65535] of byte;
 Pmidi_device_t = ^midi_device_t;
 midi_device_t = record
	user_device_name:string;				// User-defined name for this device
	enabled:boolean;						// Is this device currently enabled for playback?
	usable:boolean;						// Does the user consider this device usable?
	standards:integer;						// Device type flags (MT-32, GM, GS, XG)
	dx_device:boolean;						// Is this a DirectMusic device?
	input_device:boolean;					// Is this an input device?
	incaps:MIDIINCAPS;					// Input device caps
	outcaps:MIDIOUTCAPS;				// Output device caps
	next:Pmidi_device_t;
 end;

 channel_state_t = record
	// State values
	normal_program:byte;					// Normal program used by this channel
	override_program:byte;				// Program that the user has overridden this channel with
	notes:packed array[0..127] of byte;			// Current velocity for each possible note (0 is off)
	controllers:array[0..127] of shortint;	// Continuous controller values (-1 is unset)
	controller_overridden:array[0..127] of {shortint}boolean;	// Is this controller overridden with a user-specified value?
	note_count:integer;						// Number of notes currently on
	displayed_controller:shortint;			// Currently displayed controller (over background for note display)
	lock_controller:boolean;		// Prevent automatically switching displayed controller to last used?
	// State flags
	program_overridden:boolean;				// Has the user overridden this channel's program?
	muted:boolean;							// Is this channel muted?
	drawn:boolean;							// Have lines been drawn for this channel since last frame?
	// Historical values
	track_references:array[0..MAX_MIDI_TRACKS-1] of shortint;	// Tracks which have events on this channel
	last_note_pitch:shortint;		// Last note pitch played from this track
	last_note_velocity:shortint;		// Last note velocity played from this track
	last_controller:shortint;		// Last controller set from this track
	last_controller_value:shortint;	// Last controller value set from this track
	last_program:shortint;			// Last program set from this track
	last_bank:shortint;				// Last bank set from this track
	last_pitch_bend:integer;			// Last pitch bend set from this track
 end;

 midi_state_t = record
	perform_analysis:boolean;				// Should this song be analyzed on loading?
	analyzing:boolean;						// Is this song currently being analyzed?
	seek_sliding:boolean;					// Is the seek bar sliding? (user holding down button)
	seeking:boolean;						// Is this song currently seeking?
	seek_to:double;						// Place to seek to (in milliseconds)
	stop_requested:boolean;					// Has there been a request to stop the playback thread?
	playing:boolean;						// Is the playback thread playing?
	paused:boolean;							// Is the playback thread paused?
	found_note_on:boolean;					// Has there been a Note On event yet?
	first_note_on:double;				// Time when the first Note On event was found
	finished_naturally:boolean;				// Did the current MIDI file finish naturally?
	loop_count:integer;						// Number of times left to loop the playback
	starttime:double;					// Starting playback time
	curtime:double;						// Current playback time
	tempo:longword;					// Current tempo
	tick_length:double;					// Current tick length in milliseconds
	song_length:integer;					// Length of song in seconds
	num_events:integer;						// Number of MIDI events in song
	peak_polyphony:integer;					// Peak polyphony seen throughout the song
	uses_percussion:boolean;				// Does this MIDI file use percussion? (channel 10)
	highest_pitch_bend:integer;				// Highest pitch bend used in song (used in pitch bend visualization calculations)
	mod_velocity:integer;			// Velocity modification (-127 to 127)
	mod_pitch:integer;				// Pitch modification (-127 to 127)
	current_text:array[0..255] of char;				// Current text event being displayed
	title_displayed:boolean;				// Has the title been displayed yet?
	channels:array[0..15] of channel_state_t;		// Channel state structs
	midi_standard:integer;					// MIDI standard used by this MIDI file (as sysex indicates)
	filename:string;
 end;

 midi_header_t = packed record
	id:longword;							// MIDI header ID ("MThd")
	header_size:longword;			// Size of header in bytes
	file_format:word;			// MIDI file type (0, 1, or 2)
	num_tracks:word;			// Number of tracks in MIDI file
	num_ticks:word;			// Number of ticks per quarter note (used to calc tempo)
 end;

 track_header_t = record
	id:longword;							// Track header ID ("MTrk")
	length:longword;				// Length of track chunk in bytes
	num_events:longword;			// Number of events in this track
	num_events_done:longword;		// Number of events processed in this track
	data:PByteArray;				// Pointer to beginning of track chunk
	dataptr:PByteArray;				// Pointer to current location in track chunk
	start:double;						// Starting time of current event
	trigger:double;						// Ending time of current event
	dt:double;							// Length of current event in ticks
	enabled:boolean;						// Is this track enabled?
	tracknum:word;			// Number of this track
	lastcmd:byte;				// Last command byte (used in running mode)
	// Historical values
	name:array[0..255] of char;						// Name of this track
	last_channel:shortint;			// Last channel used by this track
	last_note_pitch:shortint;		// Last note pitch played from this track
	last_note_velocity:shortint;		// Last note velocity played from this track
	last_controller:shortint;		// Last controller set from this track
	last_controller_value:shortint;	// Last controller value set from this track
	last_program:shortint;			// Last program set from this track
	last_bank:shortint;				// Last bank set from this track
	last_pitch_bend:integer;			// Last pitch bend set from this track
 end;

 Pmidi_text_t = ^midi_text_t;
 midi_text_t = record
	midi_text:string;					// MIDI text string
	text_type:integer;						// Type of text event
	midi_time:double;					// Time (in seconds) when this text event occurs
	track:integer;							// Track this text is embedded in
	track_offset:integer;					// Offset (in bytes) where this text event occurs
	next:Pmidi_text_t;
 end;

 Pmidi_sysex_t = ^midi_sysex_t;
 midi_sysex_t = record
	data:PByteArray;				// Pointer to sysex data (in track chunk in memory)
	length:integer;							// Length of this sysex
	midi_time:double;					// Time (in seconds) when this sysex occurs
	track:integer;							// Track this sysex is embedded in
	track_offset:integer;					// Offset (in bytes) where this sysex occurs
	channel:integer;						// Channel this sysex data goes out on
	next:Pmidi_sysex_t;
 end;

procedure MIDIEnumDevices(cb:TComboBox);
procedure init_midi_out(d:integer);
procedure close_midi_out;
function load_midi(filename:string):boolean;
procedure all_notes_off_channel(channel:integer);
procedure all_notes_off;
function process_midi_event(var th:track_header_t):boolean;
procedure set_tempo(new_tempo:integer);
procedure note_on(on_:boolean;note,velocity,chan:byte);
procedure set_channel_program(channel,prog,bank:integer);
procedure output_sysex_data(channel:byte; data:PByteArray; length:integer);
procedure update_note_volume(channel, note, volume:byte);
procedure set_channel_mute(channel:integer;mute:boolean);
procedure init_mt32_state;
function MIDIThreadFunc(a:pointer):dword;stdcall;
procedure StopMIDIThread;

var
 midi_devices:Pmidi_device_t = nil;		// MIDI devices
 hout:HMIDIOUT = 0;
 ms:midi_state_t;						// MIDI state struct
 mh:midi_header_t;						// MIDI file header struct
 th:array[0..MAX_MIDI_TRACKS-1] of track_header_t;		// Track state structs
 midi_text_events:Pmidi_text_t = nil;	// MIDI text events
 midi_sysex_events:Pmidi_sysex_t = nil;	// MIDI sysex events
 mt32_patch_groups:array[0..127] of byte;			// MT-32 timbre group for a given patch
 mt32_patch_programs:array[0..127] of byte;			// MT-32 program number for a given patch
 mt32_memory_names:array[0..63] of array [0..10] of char;			// MT-32 memory timbre names

// Timing variable
PeriodMin:longword = 1;

MIDIThreadID:DWORD;
MIDIThreadH:THandle = 0;

MIDILoop:boolean = False;

implementation

uses m1;

var
	infile:file;
  last_tempo:integer = 0;
  sysex_data_buf:array of byte;
//	interpret_sysex_buf:array[0..255] of byte;
//  timbreName:array[0..15] of byte;

function IntelWord(Wrd:word):word;
asm
xchg al,ah
end;

function IntelDWord(DWrd:dword):dword;
asm
xchg al,ah
ror eax,16
xchg al,ah
end;

function MAKEBYTE(a,b:byte):byte;
begin
Result := (a and 15) or (b shl 4)
end;

procedure MIDIEnumDevices(cb:TComboBox);
var
 outdevs:integer;
 outcaps:MIDIOUTCAPS;
 i,ret:integer;
 dev,ld:Pmidi_device_t;
begin
 ld := nil;
 cb.Items.Add('MIDI Mapper');
 outdevs := midiOutGetNumDevs;
 for i := 0 to outdevs - 1 do
	begin
		ret := midiOutGetDevCaps(i, @outcaps, sizeof(outcaps));
		if (ret <> MMSYSERR_NOERROR) then	continue;

		// Set up a device node for this device
    New(dev);
		dev.user_device_name := outcaps.szPname;
		Move(outcaps, dev.outcaps, sizeof(MIDIOUTCAPS));
		dev.usable := True;
		dev.standards := MIDI_STANDARD_GM;
		dev.next := nil;
		if ld = nil then
			midi_devices := dev
		else
			ld.next := dev;
		ld := dev;
    cb.Items.Add(outcaps.szPname)
  end;
cb.ItemIndex := 0
end;

procedure init_midi_out(d:integer);
begin

	if hout <> 0 then
	 begin
		midiOutClose(hout);
		hout := 0;
	 end;

	if midiOutOpen(@hout, d, 0, 0, 0) <> MMSYSERR_NOERROR then
	 begin
		hout := 0;
		MessageBox(0,'Unable to open MIDI-out device:', 'midiOutOpen failed...', MB_ICONERROR);
   end;
end;

procedure close_midi_out;
var
 i:integer;
begin

	if hout <> 0 then
   begin
		all_notes_off;
		midiOutReset(hout);
    i := 0;
		while (midiOutClose(hout) <> MMSYSERR_NOERROR) and (i < 10) do
     begin
      inc(i);
			Sleep(200);
     end;
		if i = 10 then
			MessageBox(0, 'Unable to close MIDI-out device!', 'DMIDI Error', MB_ICONERROR);
		hout := 0;
   end;
end;

// Turns all notes off on a specific channel
procedure all_notes_off_channel(channel:integer);
begin
	// Clear saved note velocities for this channel
  FillChar(ms.channels[channel].notes,128,0);
	ms.channels[channel].note_count := 0;
	// Turn off notes on the MIDI-out device
	if hout <> 0 then
	 begin
		// Better safe than sorry!
		midiOutShortMsg(hout, MAKELONG(MAKEWORD($B0 + channel, 120), MAKEWORD(0, 0)));
		midiOutShortMsg(hout, MAKELONG(MAKEWORD($B0 + channel, 121), MAKEWORD(0, 0)));
		midiOutShortMsg(hout, MAKELONG(MAKEWORD($B0 + channel, 123), MAKEWORD(0, 0)));
	 end;
end;

// Turns all notes off on all channels
procedure all_notes_off;
var
 c:integer;
begin
	for c := 0 to 15 do
   all_notes_off_channel(c);
end;

// Adds a new node to the list of MIDI text events
function new_midi_text(midi_text:string; text_type:integer; midi_time:double; track, track_offset:integer):Pmidi_text_t;
var
	p,n:Pmidi_text_t;
	ch:integer;
begin

	// Allocate memory for the new node
  new(n);

	// If the list is empty.. add this as the top node
	if midi_text_events = nil then
		midi_text_events := n
	else
	 begin
		// Otherwise find the end of the list and add this node to the end
    p := midi_text_events; while p.next <> nil do p := p.next;
		p.next := n;
	 end;

	// Initialize data for this node
	n.text_type := text_type;
	n.midi_time := midi_time;
	n.track := track;
	n.track_offset := track_offset;
	n.next := nil;

	if midi_text <> '' then
	 begin
		// Sneaky trick to detect MT-32 MIDI files!
		if (ms.midi_standard = MIDI_STANDARD_NONE) and (midi_text = 'MT-32') then
			ms.midi_standard := MIDI_STANDARD_MT32;

		// Cool trick to break up lines with CR/LF pairs
		// Search for a CR or LF
		ch := Pos(midi_text, #13);
		if ch = 0 then ch := Pos(midi_text, #10);
		// Found one?  Good.. let's break up the line into multiple text entries!
		if ch <> 0 then
		 begin
			// Terminate the string at this character and move to the next character
			midi_text[ch] := #0; inc(ch);
			// See if there's another CR/LF to seek past
			if midi_text[ch] in [#13,#10] then inc(ch);
			// Make a new text entry with what we just seeked past
			new_midi_text(PChar(@midi_text[ch]), text_type, midi_time, track, track_offset);
		 end;

		n.midi_text := PChar(midi_text)
	 end;

	// Return a pointer to the new node
	Result := n;
end;

// Removes all nodes in the list of MIDI text events
procedure kill_all_midi_text;
var
 p,next:Pmidi_text_t;
begin
	p := midi_text_events;
	while p <> nil do
   begin
		next := p.next;
		Dispose(p);
		p := next;
	 end;
	midi_text_events := nil;
end;

procedure check_midi_standard(data:PByteArray);
begin
	case data[0] of								// Manufacturer ID
		$41:	// Roland
			case data[2] of
				$16:	// MT-32
					ms.midi_standard := MIDI_STANDARD_MT32;
				$42:	// GS Message
					ms.midi_standard := MIDI_STANDARD_GS;
      end;
		$43:	// Yamaha
			case data[2] of
				$4C:	// XG Message
					ms.midi_standard := MIDI_STANDARD_XG;
			end;
		$7E,	// General MIDI
		$7F:
			ms.midi_standard := MIDI_STANDARD_GM;
  end;
end;

// Adds a new node to the list of MIDI sysex events
function new_midi_sysex(data:PByteArray; length:integer; midi_time:double; track,track_offset,channel:integer):Pmidi_sysex_t;
var
	p,n:Pmidi_sysex_t;
begin

	// Allocate memory for the new node
  new(n);

	// If the list is empty.. add this as the top node
	if midi_sysex_events = nil then
		midi_sysex_events := n
	else
	 begin
		// Otherwise find the end of the list and add this node to the end
    p := midi_sysex_events; while p.next <> nil do p := p.next;
		p.next := n;
	 end;

	// See if the data is a MIDI reset
	check_midi_standard(data);

	// Initialize data for this node
	n.data := data;
	n.length := length;
	n.midi_time := midi_time;
	n.track := track;
	n.track_offset := track_offset;
	n.channel := channel;
	n.next := nil;

	// Return a pointer to the new node
	Result := n;
end;

// Removes all nodes in the list of MIDI sysex events
procedure kill_all_midi_sysex;
var
	p,next:Pmidi_sysex_t;
begin
	p := midi_sysex_events;
	while p <> nil do
	 begin
		next := p.next;
		Dispose(p);
		p := next;
   end;
	midi_sysex_events := nil;
end;

function read_short:word;
var
 v:word;
begin
	BlockRead(infile,v,2);
	Result := IntelWord(v)
end;

function read_int:longword;
var
 v:longword;
begin
	BlockRead(infile,v,4);
	Result := IntelDWord(v)
end;

function read_bytes_mem(var th:track_header_t; buf:PChar; num:integer):integer;
begin
  Move(th.dataptr^,buf^,num);
	inc(integer(th.dataptr),num);
	Result := num
end;

function read_byte_mem(var th:track_header_t):byte;
begin
  Result := th.dataptr^[0];
  inc(integer(th.dataptr));
end;

function read_int_mem(var th:track_header_t):longword;
begin
	Result := IntelDWord(PLongWord(th.dataptr)^);
  inc(integer(th.dataptr),4);
end;

function read_short_mem(var th:track_header_t):word;
begin
	Result := IntelWord(PWord(th.dataptr)^);
  inc(integer(th.dataptr),2);
end;

function read_vlq_mem(var th:track_header_t):longword;
var
 value:longword;
 c:byte;
 p:PByteArray;
begin
	p := th.dataptr;

    value := p^[0];
    inc(integer(p));
    if (value and $80) <> 0 then
     begin
        value := value and $7f;
        repeat
         c := p^[0];
         inc(integer(p));
    		 value := (value shl 7) + (c and $7f);
        until (c and $80) = 0;
     end;
	th.dataptr := p;
  Result := value;
end;

function analyze_midi:boolean;
var
 tracks,i,j,num_events:integer;
 curtime,starttime,nexttrigger:double;
// p:Pmidi_text_t;
 first_pass:boolean;
begin
	tracks := mh.num_tracks;
	num_events := 0;

	// Initialize playback parameters
	ms.midi_standard := MIDI_STANDARD_NONE;
	ms.stop_requested := False;
	ms.tempo := 500000 div 1000;
	ms.tick_length := ms.tempo / mh.num_ticks;
	ms.found_note_on := False;
	ms.first_note_on := 0.0;
	ms.analyzing := True;

	// Initialize track parameters
	for i := 0 to tracks - 1 do
	 begin
		th[i].name[0] := #0;
		th[i].num_events := 0;
		th[i].num_events_done := 0;
		th[i].last_channel := -1;
		th[i].last_controller := -1;
		th[i].last_controller_value := -1;
		th[i].last_note_pitch := -1;
		th[i].last_note_velocity := -1;
		th[i].last_program := -1;
		th[i].last_bank := -1;
		th[i].lastcmd := 0;
		th[i].enabled := True;
		th[i].trigger := 0.0;
		th[i].dataptr := th[i].data;
		th[i].tracknum := i;
	 end;

	// Initialize historical channel values
	for i := 0 to 15 do
	 begin
		for j := 0 to MAX_MIDI_TRACKS - 1 do
			ms.channels[i].track_references[j] := -1;
		ms.channels[i].last_controller := -1;
		ms.channels[i].last_controller_value := -1;
		ms.channels[i].last_note_pitch := -1;
		ms.channels[i].last_note_velocity := -1;
		ms.channels[i].last_program := -1;
		ms.channels[i].last_bank := -1;
	 end;

	// Record starting time
  starttime := 0.0;
	curtime := 0.0;
	ms.starttime := starttime;

	// Loop until we've been asked to stop
	first_pass := True;
	while not ms.stop_requested do
	 begin
		// Schedule the next "wakeup time" MAX_MIDI_WAIT milliseconds into the future
		nexttrigger := curtime + MAX_MIDI_WAIT;
		// See if any tracks have pending data
		ms.stop_requested := True;
		for i := 0 to tracks - 1 do
     begin
			// Continue on to the next track if this track is disabled
			if not th[i].enabled then
				continue
			else
				ms.stop_requested := False;

			// Read the next event's delta time
			if (first_pass) then
			 begin
				th[i].dt := read_vlq_mem(th[i]) * ms.tick_length;
				th[i].trigger := curtime + th[i].dt;
			 end;

			// Process MIDI events until one is scheduled for a time in the future
			while (curtime >= th[i].trigger) do
			 begin
				// Process the event for this track
				ms.curtime := curtime;
				if (process_midi_event(th[i])) then
         begin
          Result := True;
					exit;
         end;
				inc(num_events);
				inc(th[i].num_events);

				// Check for end of track
				if not th[i].enabled then	break;

				if longword(th[i].dataptr) >= (longword(th[i].data) + th[i].length) then
				 begin
					th[i].enabled := false;
					break;
				 end;

				// Read the next event's delta time
				th[i].dt := read_vlq_mem(th[i]) * ms.tick_length;
				th[i].trigger := th[i].trigger + th[i].dt;
			 end;
			// Check for end of track
			if not th[i].enabled then	continue;

			// See if this track's trigger is the more recent than the next scheduled trigger
			// If so, make it the next trigger
			if (th[i].trigger < nexttrigger) then
				nexttrigger := th[i].trigger;
		 end;
		first_pass := False;
		curtime := nexttrigger;
	 end;
	// Analysis complete

	// Store song length and # of events
	ms.song_length := round(curtime / 1000.0);
	ms.num_events := num_events;
	// Set up velocity mod
  ms.mod_velocity := 0;
	// Set up pitch mod
	if ms.mod_pitch <> 0 then all_notes_off;
	ms.mod_pitch := 0;
	// Set song length text
  Form1.Label4.Caption := Format('%d:%02d', [ms.song_length div 60, ms.song_length mod 60]);
	// Set up song position slider
  Form1.ScrollBar1.Position := 0;
  Form1.ScrollBar1.Max := ms.song_length;

	Result := False;
end;

function load_midi(filename:string):boolean;
var
  outfile:text;
	id:longword;
	s:string;
  buf:array[0..1023] of byte;
  track,i,len:longword;
  success:boolean;

begin
  Result := False;

	track := 0;
	success := False;

	// Save the given filename
  ms.filename := filename;

	// Dump any previously-loaded MIDI text and sysex events
	kill_all_midi_text;
	kill_all_midi_sysex;

	// Unload any previously-loaded MIDI data
	if mh.num_tracks <> 0 then
		for i := 0 to mh.num_tracks - 1 do
		 begin
			th[i].dataptr := nil;
			if th[i].data <> nil then FreeMem(th[i].data);
      FillChar(th[i],sizeof(track_header_t),0);
		 end;

	// Reset percussion flag
	ms.uses_percussion := False;

	// Reset highest pitch bend value to zero
	ms.highest_pitch_bend := 0;

	// Reset channel states
	for i := 0 to 15 do
   begin
		ms.channels[i].program_overridden := False;
		ms.channels[i].muted := False;
	 end;

	// Decide whether or not to perform pre-analysis
	ms.perform_analysis := False;

	if ms.perform_analysis then
   begin
		AssignFile(outfile,'c:\miditest.txt'{analysis_file});
    Rewrite(outfile);
   end;
  FileMode := 0;
	AssignFile(infile,filename);
  Reset(infile,1);

	try
	if ms.perform_analysis then
		Writeln(outfile,'Analysis of ',filename);

	// Read the MIDI file header
	BlockRead(infile, id, 4);

	// Check to see if the "MThd" ID is correct
	if MThd <> id then
	 begin
		// No?  Check for a RIFF header (RMID format)
		if RIFF = id then
		 begin
			// Yes!  Ok, skip past 4 byte filesize + 4 byte RMID
			BlockRead(infile, id, 4);
			BlockRead(infile, id, 4);
			BlockRead(infile, id, 4);
			BlockRead(infile, id, 4);
			// Now read the MIDI file header
			BlockRead(infile, id, 4);
			// Now look for MThd
			if MThd <> id then
			 begin
        s := filename + ' is not a MIDI file because it does not begin with "MThd".';
				if ms.perform_analysis then	Writeln(outfile,s);
				MessageBox(0, PChar(s), 'DMIDI Error', MB_ICONERROR);
        exit;
			 end
     end
		else
		 begin
			// STILL no MThd?  Search through the beginning of the file for MThd
			Seek(infile,0);
			BlockRead(infile,buf,1024,len);
      i := 0;
			if len > 4 then
       begin
				while i < len - 4 do
         begin
					if MThd = PLongWord(buf[i])^ then break;
          inc(i)
         end;
       end;
			if (len <= 4) or (i = len - 4) then
			 begin
        s := filename + ' is not a MIDI file because it does not begin with "MThd".';
				if ms.perform_analysis then	Writeln(outfile,s);
				MessageBox(0, PChar(s), 'DMIDI Error', MB_ICONERROR);
        exit;
			 end
			else
			 begin
				Seek(infile, i);
				BlockRead(infile,id,4);
				if ms.perform_analysis then
					Writeln(outfile, 'Warning: File is damaged... ',i,
                  ' bytes of unexpected data before MThd header.');
			 end
		 end
	 end;

	// Read the header size
	mh.header_size := read_int;
	if ms.perform_analysis then
		Writeln(outfile, 'Header size: ',mh.header_size,' bytes');

	// Identify the file format
	mh.file_format := read_short;
	if ms.perform_analysis then
	 begin
		Write(outfile, 'File format: ');
		case mh.file_format of
		  0: Writeln(outfile, 'Type 0: Single-track');
			1: Writeln(outfile, 'Type 1: Multiple tracks, synchronous');
			2: Writeln(outfile, 'Type 2: Multiple tracks, asynchronous');
    else Writeln(outfile, 'Unknown (', mh.file_format, ')');
    end
   end;
  if not mh.file_format in [0..2] then
   begin
		MessageBox(0,PChar('Unknown MIDI file type: '
                      + IntToStr(mh.file_format)), 'DMIDI Error', MB_ICONERROR);
		exit
   end;

	// Output other information present in the header
	mh.num_tracks := read_short;
	mh.num_ticks := read_short;
	if ms.perform_analysis then
	 begin
		Writeln(outfile, 'Number of tracks: ', mh.num_tracks);
		Writeln(outfile, 'Ticks per quarter note: ', mh.num_ticks);
	 end;

	// Read tracks
	while not eof(infile) do
	 begin
		// Read the track header
		BlockRead(infile, id, 4);
		// Check to see if the "MTrk" ID is correct
		if MTrk <> id then
		 begin
      s := 'Track header ' + IntToStr(track + 1) + ' is not correct because it does not begin with "MTrk".';
			if ms.perform_analysis then Writeln(outfile, s);
			if track < mh.num_tracks then	MessageBox(0, PChar(s), 'DMIDI Warning', MB_ICONWARNING);
			break;
     end;
		// Read track length in bytes
		th[track].length := read_int;

		if ms.perform_analysis then
     begin
      Writeln(outfile);
			Writeln(outfile, '---- Track ',track + 1,' (',th[track].length,' bytes) ----');
      Writeln(outfile);
     end;

		// Read the track data into memory
    GetMem(th[track].data,th[track].length);
		if th[track].data = nil then
		 begin
			MessageBox(0, 'GetMem failed trying to allocate memory for track data', 'DMIDI Error', MB_ICONERROR);
			break;
		 end;
		BlockRead(infile,th[track].data^,th[track].length);
		inc(track);
		if (track >= MAX_MIDI_TRACKS) then
		 begin
			MessageBox(0, PChar('Warning: Max number of MIDI tracks reached (' +
           IntToStr(MAX_MIDI_TRACKS) +
        ').'#13#10'I refuse to read any more from this file.'), 'DMIDI Warning', MB_ICONWARNING);
			break;
		 end;
	 end;
	success := True;

	finally

		mh.num_tracks := track;
		CloseFile(infile);
		if ms.perform_analysis then CloseFile(outfile);
		if (success) and not analyze_midi then
     Result := True
		else
     MessageBox(0,PChar(filename + ' file load error'),'Error',MB_ICONERROR);
	end;

end;

function process_midi_event(var th:track_header_t):boolean;
var
 text:PChar;
 id:longword;
// eventname:array[0..127] of char;
 i, len:longword;
 cmd, d1, d2:byte;
 intd:longword;
 pitchbend:word;
 channel:integer;
 cmdptr:PByteArray;
 sysexptr:PByteArray; // Pointer to beginning of sysex data
begin
	cmdptr := th.dataptr;		// Pointer to the beginning of this event

	// Read the MIDI event command
	cmd := read_byte_mem(th);

	// Handle running mode
	if (cmd < 128) then
	 begin
		dec(integer(th.dataptr));
		cmd := th.lastcmd;
	 end
	else
		th.lastcmd := cmd;

	if (cmd = $FF) then		// Meta-event
	 begin
		cmd := read_byte_mem(th);
		len := read_vlq_mem(th);
		if len > th.length - (longword(th.dataptr) - longword(th.data)) then
		 begin
			MessageBox(0, 'Meta-event length error', 'DMIDI Read Error', MB_ICONERROR);
			Result := True;
      exit
		 end;
		text := nil;
		case (cmd) of

			$00:		// Sequence number
				read_short_mem(th);
			$01,		// Text
			$02,		// Copyright info
			$03,		// Sequence or track name
			$04,		// Track instrument name
			$05,		// Lyric
			$06,		// Marker
			$07:		// Cue point
       begin
//        midi_text_event_descriptions[cmd]
				// Read the text
				GetMem(text,len + 1);
				read_bytes_mem(th, text, len);
				text[len] := #0;


				// If it's a track name, save it
				if cmd in [3,4] then
				 begin
					strlcopy(th.name, text, sizeof(th.name));
					th.name[sizeof(th.name) - 1] := #0;
				 end
				else
					if (th.name[0] = #0) then
					 begin
  					strlcopy(th.name, text, sizeof(th.name));
	  				th.name[sizeof(th.name) - 1] := #0;
					 end;

				// Save the text in the list of text events
				if (ms.analyzing and not ms.seeking) then
					new_midi_text(text, cmd, (ms.curtime - ms.starttime) / 1000, th.tracknum, longword(cmdptr) - longword(th.data));
  		 end;
			$20:		// MIDI Channel prefix
				read_byte_mem(th);
			$21:		// MIDI Port
				read_byte_mem(th);
			$2F:		// End of track
				th.enabled := False;
			$51:		// Set tempo
       begin
				read_bytes_mem(th, @id, 3);
        intd := IntelDWord(id shl 8);
				set_tempo(intd);
			 end;
			$54:		// SMPTE Offset
       begin
				read_byte_mem(th);
				read_byte_mem(th);
				read_byte_mem(th);
				read_byte_mem(th);
				read_byte_mem(th);
       end;
			$58:		// Time signature
       begin
				read_byte_mem(th);
				read_byte_mem(th);
				read_byte_mem(th);
				read_byte_mem(th);
       end;
			$59:		// Key signature
       begin
				read_byte_mem(th);
				read_byte_mem(th);
			 end;
			$7F:		// Sequencer-specific information
       for i := 0 to len - 1 do
			  read_byte_mem(th);
			else		// Unknown
       for i := 0 to len - 1 do
 				read_byte_mem(th);
		end;
		if text <> nil then	FreeMem(text);
	 end
	else
	 begin
		channel := cmd and 15;
		// *** cool symmetric part follows: ***
		// Save this channel number in the track struct for historical purposes
		th.last_channel := channel;
		// Save this track number in the channel struct for historical purposes
		ms.channels[channel].track_references[th.tracknum] := 1;
		case cmd shr 4 of	// Normal event

			$08: // Note off
       begin
				d1 := read_byte_mem(th);
				d2 := read_byte_mem(th);
				if not ms.analyzing then
         note_on(FALSE, d1, d2, channel);
			 end;
			$09: // Note on
       begin
				d1 := read_byte_mem(th);
				d2 := read_byte_mem(th);
				if not ms.analyzing then
				 begin
					th.last_note_pitch := d1;
					th.last_note_velocity := d2;
					note_on(TRUE, d1, d2, channel);
				 end;
				if not ms.found_note_on then
				 begin
					ms.found_note_on := True;
					ms.first_note_on := ms.curtime;
				 end;
			 end;
			$0A: // Key after-touch
       begin
				d1 := read_byte_mem(th);
				d2 := read_byte_mem(th);
				if not ms.analyzing then
         midiOutShortMsg(hout, MAKELONG(MAKEWORD(cmd, d1), MAKEWORD(d2, 0)));
			 end;
			$0B: // Control Change
       begin
				d1 := read_byte_mem(th);
        {0..127 -  ?}
				d2 := read_byte_mem(th);
				// Save the controller value
				if ms.seeking then
				 ms.channels[channel].controllers[d1] := d2;
				if not ms.analyzing then
				 begin
          th.last_controller := d1;
					ms.channels[channel].last_controller := d1;
          th.last_controller_value := d2;
					ms.channels[channel].last_controller_value := d2;
					if d1 = 0 then
           begin
            th.last_bank := d2;
						ms.channels[channel].last_bank := d2;
           end;
					if not ms.channels[channel].lock_controller then
						ms.channels[channel].displayed_controller := d1;		// Automatically select which controller to display
					if not ms.channels[channel].controller_overridden[d1] then
					 begin
						ms.channels[channel].controllers[d1] := d2;			// Update controller value
						midiOutShortMsg(hout, MAKELONG(MAKEWORD(cmd, d1), MAKEWORD(d2, 0)));
           end;
				 end;
			 end;
			$0C: // Program Change
       begin
				d1 := read_byte_mem(th);
				if not ms.analyzing or ms.seeking then
				 begin
          th.last_program := d1;
					ms.channels[channel].last_program := d1;
					ms.channels[channel].normal_program := d1;
					if not ms.channels[channel].program_overridden then
           set_channel_program(channel, d1, ms.channels[channel].last_bank);
				 end;
			 end;
			$0D: // Channel after-touch
       begin
				d1 := read_byte_mem(th);
				if not ms.analyzing then
				 midiOutShortMsg(hout, MAKELONG(MAKEWORD(cmd, d1), MAKEWORD(0, 0)));
       end;
			$0E: // Pitch wheel
       begin
				intd := read_short_mem(th);
				d2 := LOBYTE(intd);
				d1 := HIBYTE(intd);
				pitchbend := d2;
				pitchbend := pitchbend shl 7;
				pitchbend := pitchbend or d1;
				if not ms.analyzing then
				 begin
					midiOutShortMsg(hout, MAKELONG(MAKEWORD(cmd, d1), MAKEWORD(d2, 0)));
          th.last_pitch_bend := integer(pitchbend) - MAX_PITCH_BEND;
					ms.channels[channel].last_pitch_bend := integer(pitchbend) - MAX_PITCH_BEND;
				 end
				else
				 begin
					// Keep track of the highest pitch bend value used in this song
					if (abs(integer(pitchbend) - MAX_PITCH_BEND) > ms.highest_pitch_bend) then
						ms.highest_pitch_bend := abs(integer(pitchbend) - MAX_PITCH_BEND);
				 end;
			 end;
			$0F: // System message
       begin
				case cmd and 15 of

					$02:
           begin
            //Song position
						read_byte_mem(th);
						read_byte_mem(th);
            //Song select
						read_byte_mem(th); //$03
            //Tune request $06
           end;
					$03:
           //Song select
           read_byte_mem(th);
//					$06: //Tune request
					$00,
					$07:	// SYSEX data
           begin
						len := read_vlq_mem(th);
						if len > th.length - (longword(th.dataptr) - longword(th.data)) then
						 begin
							MessageBox(0, 'Sysex len error', 'DMIDI Read Error', MB_ICONERROR);
							Result := True;
              exit;
						 end;
						sysexptr := th.dataptr;
						if ms.analyzing and not ms.seeking then
							new_midi_sysex(sysexptr, len, (ms.curtime - ms.starttime) / 1000.0, th.tracknum, longword(cmdptr) - longword(th.data), channel);
						inc(integer(th.dataptr),len);
						if not ms.analyzing then
						 begin
							// Interpret the sysex data, in case we need it (like MT-32)
// ?							interpret_sysex(sysexptr, len);
							// Don't send sysex when seeking on the MT-32
							if not ((ms.midi_standard = MIDI_STANDARD_MT32) and ms.seeking) then
								output_sysex_data(channel, sysexptr, len);
						 end;
					 end;
					$08:
           begin
						//fprintf(outfile, "System Message: Timing clock used when synchronization is required.\n");
					 end;
					$0A:
           begin
						//fprintf(outfile, "System Message: Start current sequence\n");
					 end;
					$0B:
           begin
						//fprintf(outfile, "System Message: Continue a stopped sequence where left off\n");
           end;
					$0C:
           begin
						//fprintf(outfile, "System Message: Stop a sequence\n");
					 end;
					//default:
						//fprintf(outfile, "System Message: Unknown (%d)\n", LONYBBLE(cmd));
				end;
			 end;
			else
       begin
				//fprintf(outfile, "Unknown Message (%d)\n", HINYBBLE(cmd));
				read_byte_mem(th);
       end;
	  end;
	 end;

	Result := False;
end;

procedure set_tempo(new_tempo:integer);
var
 old_tick_length:double;
 i:integer;
begin

	// Calculate new tick length
	old_tick_length := ms.tick_length;
	ms.tick_length := ((new_tempo / 1000) / mh.num_ticks);

	if (new_tempo = last_tempo) then exit;

	last_tempo := new_tempo;

	// Correct outstanding track triggers on each track
	for i := 0 to mh.num_tracks - 1 do
	 begin
		if not th[i].enabled then	continue;
		th[i].dt := (th[i].trigger - ms.curtime) / old_tick_length * ms.tick_length;
		th[i].trigger := ms.curtime + th[i].dt;
	 end;
end;

procedure note_on(on_:boolean;note,velocity,chan:byte);
var
 dwParam1:DWORD;
 i:integer;
 channel:byte;
begin
	channel := chan;

	// Perform global velocity modulation
	if (velocity <> 0) and (ms.mod_velocity <> 0) then
	 begin
		i := velocity;
		inc(i,ms.mod_velocity);
		if (i < 0) then
			i := 0;
		if (i > 127) then
			i := 127;
		velocity := i;
	 end;
	// Perform global pitch modulation
	if (channel <> 9) and (ms.channels[channel].last_bank <> 127) then
		inc(note,ms.mod_pitch);

	// Update display/historic values
  if on_ then
   update_note_volume(channel, note, velocity)
  else
   update_note_volume(channel, note, 0);
	ms.channels[channel].last_note_pitch := note;
	ms.channels[channel].last_note_velocity := velocity;

	// Don't actually play the note if this channel is muted
	if ms.channels[channel].muted then exit;

	// If the bank for this channel is 127, treat it as a percussive channel! (XG)
	// This needs to be an option, in case someone actually has a real XG synth.
	if (ms.midi_standard = MIDI_STANDARD_XG) and (ms.channels[channel].last_bank = 127) then
		channel := 9;

  if on_ then
  	dwParam1 := MAKELONG(MAKEWORD(MAKEBYTE(channel, 9), note), MAKEWORD(velocity, 0))
  else
  	dwParam1 := MAKELONG(MAKEWORD(MAKEBYTE(channel, 8), note), MAKEWORD(velocity, 0));
	if hout <> 0 then	midiOutShortMsg(hout, dwParam1);
end;

procedure set_channel_program(channel,prog,bank:integer);
begin
  if (channel >= 16) or (prog >= 128) then
   begin
    MessageBox(0,'set_channel_program range error','Error',MB_ICONERROR);
    halt(1);
   end;

	if hout <> 0 then
		midiOutShortMsg(hout, MAKELONG(MAKEWORD(MAKEBYTE(channel, $0C), prog), MAKEWORD(0, 0)));

end;

procedure output_sysex_data(channel:byte; data:PByteArray; length:integer);
var
 mh:MIDIHDR;
begin

	if hout = 0 then exit;

	// Realloc the sysex buffer if it's too small
	if (length + 2 > System.Length(sysex_data_buf)) then
   SetLength(sysex_data_buf,length + 32);


	sysex_data_buf[0] := $F0;					// Sysex begin command
	Move(data^, sysex_data_buf[1], length);	// Copy the sysex data into the local sysex buffer

	// Prepare the MIDI out header
	FillChar(mh, sizeof(mh), 0);
	mh.lpData := @sysex_data_buf[0];
	mh.dwBufferLength := length + 1;
	mh.dwBytesRecorded := length + 1;
	// Prepare the sysex buffer for output
	midiOutPrepareHeader(hout, @mh, sizeof(mh));

	// Send the sysex buffer!
	midiOutLongMsg(hout, @mh, sizeof(mh));

	// Unprepare the sysex buffer
	midiOutUnprepareHeader(hout, @mh, sizeof(mh));
end;

// Updates the volume for a given note on a given channel
procedure update_note_volume(channel, note, volume:byte);
begin
	if volume <> 0 then
	 begin
		// Note on
		if ms.channels[channel].notes[note] = 0 then
			inc(ms.channels[channel].note_count);
	 end
	else
	 begin
		// Note off
		if ms.channels[channel].notes[note] <> 0 then
			dec(ms.channels[channel].note_count);
	 end;
	// Store the value
	ms.channels[channel].notes[note] := volume;
end;

procedure set_channel_mute(channel:integer;mute:boolean);
var
 vel,note:byte;
 was_muted:boolean;
begin

	was_muted := ms.channels[channel].muted;
	ms.channels[channel].muted := mute;

	if mute then
	 begin
		// Mute the channel
		all_notes_off_channel(channel);
	 end
	else
	 begin
		// Unmute the channel

		// Restore any notes that should be playing (unless this is the percussion channel)
		if (channel <> 9) and was_muted then
			for note := 0 to 127 do
			 begin
				vel := ms.channels[channel].notes[note];
				if vel <> 0 then
					note_on(TRUE, note, vel, channel);
			 end;
	 end;

end;

// Initializes MT-32 state values on program startup
procedure init_mt32_state;
var
 i:integer;
begin

	for i := 0 to 127 do
	 begin
		if (i < 64) then
			mt32_patch_groups[i] := 0
		else
			mt32_patch_groups[i] := 1;
		mt32_patch_programs[i] := i;
	 end;

	for i := 0 to 63 do
		mt32_memory_names[i][0] := #0;
end;

function MIDIThreadFunc(a:pointer):dword;stdcall;
var
 tracks:integer;
 i, j, polyphony{, elapsed}:integer;
 curtime, starttime, pausetime, tmptime, timediff:double;
 nexttrigger:double;
 first_pass:boolean;
 tracks_active:boolean;
 seeking:boolean;
label BeginPlayback;
begin
  Result := STILL_ACTIVE - 1;
	tracks := mh.num_tracks;
	seeking := False;

	// Initialize seeking states
	ms.seeking := False;
	ms.seek_sliding := False;
	ms.seek_to := 0.0;

	// Init MIDI-out device
	init_midi_out(integer(a));
	if hout = 0 then exit;

	// Set playback thread priority
	SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_HIGHEST);

	for i := 0 to 15 do
	 begin
		// Set overridden instrument channels
		if (ms.channels[i].program_overridden) then
		 set_channel_program(i, ms.channels[i].override_program, 0);

		// Update channel mutes
		set_channel_mute(i, ms.channels[i].muted);

		// Reset displayed controller to none
		ms.channels[i].displayed_controller := -1;

		// Reset controller locks
		ms.channels[i].lock_controller := False;
 	 end;

	// Initialize modifiers
	ms.mod_pitch := 0;
	ms.mod_velocity := 0;

	// Initialize loop count
	ms.loop_count := 1;

	while ms.loop_count > 0 do
	 begin
    dec(ms.loop_count);
BeginPlayback:
		// Initialize tempo
		set_tempo(trunc(60000000.0 / ms.tempo));

		// Initialize playback parameters
		ms.analyzing := ms.seeking;
		ms.playing := True;
		ms.paused := False;
		ms.finished_naturally := FALSE;
		ms.stop_requested := False;
		ms.tempo := trunc(500000 / 1000);
		ms.tick_length := ms.tempo / mh.num_ticks;
		ms.peak_polyphony := 0;
		ms.current_text[0] := #0;
		ms.title_displayed := False;
//		SendDlgItemMessage(hwndApp, IDC_POLYPHONY_METER, PBM_SETRANGE, 0, MAKELPARAM(0, ms.peak_polyphony));

		// Reset channel states
		for i := 0 to 15 do
		 begin
			// Reset # of notes currently being played on this channel
			ms.channels[i].note_count := 0;
			ms.channels[i].drawn := True;
			// Reset program values
			ms.channels[i].normal_program := 0;
			// Reset note values
      FillChar(ms.channels[i].notes,sizeof(ms.channels[i].notes),0);
			// Reset controller values
			for j := 0 to 127 do
			 begin
				ms.channels[i].controllers[j] := -1;
				ms.channels[i].controller_overridden[j] := False;
			 end;
			// Reset historical values
			ms.channels[i].last_controller := -1;
			ms.channels[i].last_controller_value := -1;
			ms.channels[i].last_note_pitch := -1;
			ms.channels[i].last_note_velocity := -1;
			ms.channels[i].last_pitch_bend := -1;
		 end;

		// Initialize track parameters
		for i := 0 to tracks - 1 do
		 begin
			th[i].num_events_done := 0;
			th[i].last_controller := -1;
			th[i].last_controller_value := -1;
			th[i].last_note_pitch := -1;
			th[i].last_note_velocity := -1;
			th[i].last_pitch_bend := -1;
			th[i].lastcmd := 0;
			th[i].enabled := True;
			th[i].trigger := 0.0;
			th[i].dataptr := th[i].data;
			th[i].tracknum := i;
		 end;

    //Set max MMTimers precision
    timeBeginPeriod(PeriodMin);
		// Record starting time
    curtime := timeGetTime;
		ms.starttime := curtime;
    starttime := curtime;

		// Loop until we've been asked to stop
		first_pass := True;
		while (not ms.stop_requested) do
		 begin
			// See if we've been asked to seek and we're not already doing it
			if ms.seeking and not seeking then
			 begin
				seeking := True;
				goto BeginPlayback;		// Aaaah!  A goto!!
			 end;
			// Get current time
			if (not seeking) then curtime := timeGetTime;
			// Schedule the next "wakeup time" MAX_MIDI_WAIT milliseconds into the future
			nexttrigger := curtime + MAX_MIDI_WAIT;
			// See if any tracks have pending data
			tracks_active := False;
			for i := 0 to tracks - 1 do
			 begin
				// Continue on to the next track if this track is disabled
				if (th[i].enabled) then
					tracks_active := True
				else
					continue;

				// Disable this track if its data pointer has reached or exceeded the track length
				if longword(th[i].dataptr) >= longword(th[i].data) + th[i].length then
				 begin
					th[i].enabled := FALSE;
					continue;
				 end;

				// Read the first event's delta time
				if (first_pass) then
				 begin
					th[i].dt := read_vlq_mem(th[i]) * ms.tick_length;
					th[i].trigger := curtime + th[i].dt;
				 end;

				// Process MIDI events until one is scheduled for a time in the future
				while (curtime >= th[i].trigger) do
				 begin
					// Process the event for this track
					ms.curtime := curtime;
					if process_midi_event(th[i]) then
					 begin
						ms.stop_requested := True;
						break;
					 end;
					// Increment the global number of events processed, and the number of events for this track
					inc(th[i].num_events_done);

					// Check for end of track
					if (not (th[i].enabled)) then	break;

					// Read the next event's delta time
					th[i].dt := read_vlq_mem(th[i]) * ms.tick_length;
					th[i].trigger := th[i].trigger + th[i].dt;
				 end;
				// Check for end of track
				if (not (th[i].enabled)) then continue;

				// See if this track's trigger is the more recent than the next scheduled trigger
				if (th[i].trigger < nexttrigger) then
					nexttrigger := th[i].trigger;
			 end;
			first_pass := False;
			if (not seeking) then	curtime := timeGetTime;
			// Check to see if a pause has been requested
			if (ms.paused) then
			 begin
				// Turn off the notes being played
				all_notes_off;
				// Save the time we paused at
				pausetime := curtime;
				// Wait until we're unpaused or not playing anymore
				while True do
				 begin
					Sleep(100);
					if not (ms.paused) or not (ms.playing) or ms.stop_requested then
						break;
				 end;
				// If we've been asked to stop.. break from the main playback loop
				if not (ms.playing) or ms.stop_requested then
					break;
				// We need to resume.  Correct the start and trigger times to be up-to-date!
				curtime := timeGetTime;
				starttime := starttime + curtime - pausetime;
				nexttrigger := nexttrigger + curtime - pausetime;
				for i := 0 to tracks - 1 do
					th[i].trigger := th[i].trigger + curtime - pausetime;
			 end;

			if (not seeking) then
			 begin
				// Set song current time / total time text
//				elapsed := trunc((curtime - starttime) / 1000.0);
        //   
				// Calculate and display polyphony
        polyphony := 0;
				for i := 0 to 15 do
					inc(polyphony,ms.channels[i].note_count);
					ms.peak_polyphony := polyphony;
			 end;

			// Wait until the next trigger time
			if (seeking) then
			 begin
				curtime := nexttrigger;
				// See if the seek is done
				if (seeking and (curtime - starttime >= ms.seek_to)) then
				 begin
					// Seek is done - correct timing values
					tmptime := curtime;
					curtime := timeGetTime;
					timediff := tmptime - curtime;
					for i := 0 to mh.num_tracks - 1 do
						th[i].trigger := th[i].trigger - timediff;
					// Calculate corrected start time and elapsed time
          starttime := curtime - timediff;
					ms.starttime := starttime;
//					elapsed := trunc(curtime - ms.starttime);
					ms.curtime := curtime;
					// Reset seek state flags
          ms.analyzing := False;
          ms.seeking := False;
					seeking := False;
					// Turn all notes off
					all_notes_off;

					// Set channel values
					for i := 0 to 15 do
					 begin
						// Set program changes
						if (ms.channels[i].normal_program <> 0) then
							set_channel_program(i, ms.channels[i].normal_program, ms.channels[i].last_bank);
					 end;
				 end;
			 end
			else
			 begin
				while (curtime < nexttrigger) and not (ms.stop_requested) do
				 begin
					Sleep(trunc(nexttrigger - curtime));
					curtime := timeGetTime;
         end;
				if (not tracks_active) then
				 begin
					ms.finished_naturally := TRUE;
					break;
				 end;
			 end;
		 end;
		// Break out of looping if a stop was requested
		if (ms.stop_requested) then
			break;
		// If the looping checkbox is checked, loop another time
		if (MIDILoop) then
			inc(ms.loop_count);
	 end;
	// PLAYBACK HAS STOPPED
  //Restore default MMTimers precision
  timeEndPeriod(PeriodMin);
	// Turn off all the notes, reset the MIDI-out device, and close it
	all_notes_off;
	Sleep(50);
	midiOutReset(hout);
	close_midi_out;

	// Reset the track info displays
	for i := 0 to 15 do
	 begin
    FillChar(ms.channels[i].notes,sizeof(ms.channels[i].notes),0);
		ms.channels[i].note_count := 0;
		ms.channels[i].displayed_controller := -1;
		ms.channels[i].lock_controller := False;
		ms.channels[i].drawn := True;
	 end;


	ms.playing := False;
//	PostMessage(hwndApp, WMAPP_DONE_PLAYING, 0, 0);
end;

procedure StopMIDIThread;
var
 ExCode:DWORD;
begin
if MIDIThreadH = 0 then exit;
ms.stop_requested := True;
repeat
 if not GetExitCodeThread(MIDIThreadH,ExCode) then break;
 if ExCode = STILL_ACTIVE then Sleep(1)
until ExCode <> STILL_ACTIVE;
CloseHandle(MIDIThreadH);
MIDIThreadH := 0
end;

end.
